package com.css.simulation.resource.scheduler.consumer;


import api.common.pojo.constants.DictConstants;
import api.common.pojo.dto.ProjectMessageDTO;
import api.common.util.JsonUtil;
import api.common.util.StringUtil;
import com.css.simulation.resource.scheduler.mapper.*;
import com.css.simulation.resource.scheduler.pojo.po.*;
import com.css.simulation.resource.scheduler.pojo.to.PrefixTO;
import com.css.simulation.resource.scheduler.service.ProjectService;
import com.css.simulation.resource.scheduler.util.ProjectUtil;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.kafka.annotation.KafkaListener;
import org.springframework.stereotype.Component;

import javax.annotation.Resource;
import java.util.*;

@Component
@Slf4j
public class ProjectConsumer {

    @Value("${scheduler.minio-path.project-result}")
    String resultPathMinio;
    @Value("${scheduler.host.hostname}")
    String hostname;
    @Value("${scheduler.host.username}")
    String username;
    @Value("${scheduler.host.password}")
    String password;

    // -------------------------------- Comment --------------------------------

    @Resource
    StringRedisTemplate stringRedisTemplate;
    @Resource
    ManualProjectMapper manualProjectMapper;
    @Resource
    AutoSubProjectMapper autoSubProjectMapper;
    @Resource
    VehicleMapper vehicleMapper;
    @Resource
    SensorCameraMapper sensorCameraMapper;
    @Resource
    SensorOgtMapper sensorOgtMapper;
    @Resource
    UserMapper userMapper;
    @Resource
    ClusterMapper clusterMapper;
    @Resource
    ProjectService projectService;
    @Resource
    ProjectUtil projectUtil;


    /**
     * 任务运行前首先判断用户是否拥有可分配资源
     *
     * @param projectRecord 项目启动消息
     */
    @KafkaListener(groupId = "simulation-resource-scheduler", topics = "${scheduler.start-topic}")
    @SneakyThrows
    public void cacheProject(ConsumerRecord<String, String> projectRecord) {
        String initialProjectJson = projectRecord.value();
        log.info("ProjectConsumer--cacheManualProject 接收到项目开始消息为：" + initialProjectJson);
        //1 读取 kafka 的 project 信息
        ProjectMessageDTO projectMessageDTO = JsonUtil.jsonToBean(initialProjectJson, ProjectMessageDTO.class);
        String projectId = projectMessageDTO.getProjectId();    // 手动执行项目 id 或 自动执行子项目 id
        long parallelism = projectMessageDTO.getParallelism();   // 项目并行度
        String projectType = projectMessageDTO.getType(); // 项目类型
        //2 根据 projectId 获取创建用户 id
        String userId;
        if (DictConstants.PROJECT_TYPE_MANUAL.equals(projectType)) {
            userId = manualProjectMapper.selectCreateUserById(projectId);
        } else if (DictConstants.PROJECT_TYPE_AUTO_SUB.equals(projectType)) {
            userId = autoSubProjectMapper.selectCreateUserById(projectId);
        } else {
            log.error("ProjectConsumer--cacheManualProject 项目类型错误：" + initialProjectJson);
            return;
        }
        if (StringUtil.isEmpty(userId)) {
            log.error("ProjectConsumer--cacheManualProject 未查询到项目创建人：" + initialProjectJson);
            return;
        }
        //3 获取用户类型（管理员账户、管理员子账户、普通账户、普通子账户）（独占、共享）
        UserPO userPO = userMapper.selectById(userId);
        log.info("ProjectConsumer--cacheManualProject 项目 " + projectId + " 的创建人为：" + userPO);
        String roleCode = userPO.getRoleCode();
        String useType = userPO.getUseType();
        ClusterPO clusterPO;
        if (DictConstants.ROLE_CODE_SYSADMIN.equals(roleCode) || DictConstants.ROLE_CODE_ADMIN.equals(roleCode)) {  //3-1 管理员账户和管理员子账户直接执行
            log.info("ProjectConsumer--cacheManualProject 项目 " + projectId + " 的创建人 " + userId + " 为管理员账户或管理员子账户，直接判断服务器能否执行。");
            PrefixTO redisPrefix = projectUtil.getRedisPrefixByClusterIdAndProjectId(DictConstants.SYSTEM_CLUSTER_ID, projectId);
            run(projectMessageDTO, DictConstants.SYSTEM_CLUSTER_ID, redisPrefix.getProjectRunningKey(), redisPrefix.getProjectWaitingKey());
            return;
        } else if (DictConstants.ROLE_CODE_UESR.equals(roleCode)) { //3-2 普通账户，不管是独占还是共享，都在自己的集群里排队，根据自己的独占节点排队
            clusterPO = clusterMapper.selectByUserId(userId);
            log.info("ProjectConsumer--cacheManualProject 项目 " + projectId + " 的创建人 " + userId + " 为普通账户（包括独占或共享都在自己的集群），集群为：" + clusterPO);
        } else if (DictConstants.ROLE_CODE_SUBUESR.equals(roleCode)) {
            if (DictConstants.USER_TYPE_EXCLUSIVE.equals(useType)) {   //3-3 普通子账户，根据自己的独占节点排队
                clusterPO = clusterMapper.selectByUserId(userId);
                log.info("ProjectConsumer--cacheManualProject 项目 " + projectId + " 的创建人 " + userId + " 为普通独占子账户（自己的集群），集群为：" + clusterPO);
            } else {    //3-4 共享子账户，根据父账户的共享节点排队
                String parentUserId = userPO.getCreateUserId();
                clusterPO = clusterMapper.selectByUserId(parentUserId);
                log.info("ProjectConsumer--cacheManualProject 项目 " + projectId + " 的创建人 " + userId + " 为普通共享子账户（父账户的集群），集群为：" + clusterPO);
            }
        } else {
            log.error("ProjectConsumer--cacheManualProject 项目 " + projectId + " 的创建人 " + userId + " 为未知账户类型，不予执行！");
            return;
        }
        // 获取拥有的节点数量，即仿真软件证书数量
        String clusterId = clusterPO.getId();
        int simulationLicenseNumber = clusterPO.getNumSimulationLicense();
        // 获取该集群中正在运行的项目，如果没有则立即执行
        PrefixTO redisPrefix = projectUtil.getRedisPrefixByClusterIdAndProjectId(clusterId, projectId);
        // 获取正在运行的项目的并行度总和
        int currentParallelismSum = projectUtil.getCurrentParallelismSum(redisPrefix.getClusterRunningPrefix());
        // 如果执行后的并行度总和小于最大节点数则执行，否则不执行
        if (currentParallelismSum + parallelism <= simulationLicenseNumber) {
            run(projectMessageDTO, clusterId, redisPrefix.getProjectRunningKey(), redisPrefix.getProjectWaitingKey());
        } else {
            wait(redisPrefix.getProjectWaitingKey(), projectMessageDTO);
        }
    }

    /**
     * @param projectMessageDTO 初始接收到的项目启动信息
     * @param clusterId         集群 id
     * @param projectRunningKey projectRunningKey
     * @param projectWaitingKey projectWaitingKey
     */
    public void run(ProjectMessageDTO projectMessageDTO, String clusterId, String projectRunningKey, String projectWaitingKey) {

        String projectId = projectMessageDTO.getProjectId();
        int parallelism = projectMessageDTO.getParallelism();  // 期望并行度
        //1 获取集群剩余可用并行度
        int restParallelism = projectUtil.getRestParallelism();
        //2 判断剩余可用并行度是否大于项目并行度，否则加入扩充队列
        if (restParallelism > 0L) {
            log.info("ProjectConsumer--run 集群 " + clusterId + " 执行项目 " + projectId);
            // 设置实际的并行度
            projectMessageDTO.setCurrentParallelism(Math.min(restParallelism, parallelism));   // 设置实际的并行度
            parseProject(projectMessageDTO, projectWaitingKey, projectRunningKey);
        } else {
            log.info("ProjectConsumer--cacheManualProject 服务器资源不够，项目 " + projectId + " 暂时加入等待队列。");
            wait(projectWaitingKey, projectMessageDTO);
        }
    }

    /**
     * @param projectWaitingKey 项目等待 key
     * @param projectMessageDTO 项目信息
     */
    @SneakyThrows
    public void wait(String projectWaitingKey, ProjectMessageDTO projectMessageDTO) {
        stringRedisTemplate.opsForValue().set(projectWaitingKey, JsonUtil.beanToJson(projectMessageDTO));
    }


    /**
     * @param projectMessageDTO 初始接收到的项目启动信息
     * @param projectWaitingKey projectWaitingKey
     * @param projectRunningKey projectRunningKey
     */
    @SneakyThrows
    public void parseProject(ProjectMessageDTO projectMessageDTO, String projectWaitingKey, String projectRunningKey) {
        String projectId = projectMessageDTO.getProjectId();    // 项目 id
        String projectType = projectMessageDTO.getType();   // 项目类型
        int currentParallelism = projectMessageDTO.getCurrentParallelism();   // 当前并行度
        String packageId = projectMessageDTO.getScenePackageId();   // 场景测试包 id
        long videoTime = projectMessageDTO.getMaxSimulationTime(); // 结果视频的时长
        String vehicleConfigId = projectMessageDTO.getVehicleConfigId();// 模型配置 id
        String algorithmId = projectMessageDTO.getAlgorithmId();    // 算法 id
        // -------------------------------- 0 准备 --------------------------------
        projectService.prepare(projectMessageDTO, projectWaitingKey, projectRunningKey);
        String userId = null;
        if (DictConstants.PROJECT_TYPE_MANUAL.equals(projectType)) {
            userId = manualProjectMapper.selectCreateUserById(projectId);
        } else if (DictConstants.PROJECT_TYPE_AUTO_SUB.equals(projectType)) {
            userId = autoSubProjectMapper.selectCreateUserById(projectId);
        }
        // -------------------------------- 1 查询场景 --------------------------------
        //1-1 根据场景测试包 packageId，拿到场景集合（不包括重复场景），重复场景会在发送消息时根据叶子指标发送多次。
        List<ScenePO> scenePOList = projectService.handlePackage(projectRunningKey, projectId, packageId);
        int taskTotal = scenePOList.size();
        projectMessageDTO.setTaskTotal(taskTotal);
        projectMessageDTO.setTaskCompleted(0);
        // 设置任务数量之后，获取运行节点，并将项目运行信息放入 redis
        Map<String, Integer> nodeMap;
        if (currentParallelism < taskTotal) {
            nodeMap = projectUtil.getNodeMapToUse(currentParallelism);
        } else {
            nodeMap = projectUtil.getNodeMapToUse(taskTotal);
        }
        // 将指定 node 的并行度减少
        nodeMap.keySet().forEach(nodeName -> {
            int parallelismToUse = nodeMap.get(nodeName);
            String restParallelismKey = "node:" + nodeName + ":parallelism";
            int restParallelism = Integer.parseInt(Objects.requireNonNull(stringRedisTemplate.opsForValue().get(restParallelismKey)));// 剩余可用并行度
            stringRedisTemplate.opsForValue().set(restParallelismKey, (restParallelism - parallelismToUse) + "");
        });
        // 重新设置实际使用的并行度并保存到 redis
        int realCurrentParallelism = nodeMap.values().stream().mapToInt(parallelism -> parallelism).sum();
        projectMessageDTO.setCurrentParallelism(realCurrentParallelism);
        log.info("ProjectConsume--parseProject 项目 " + projectId + " 运行在：" + nodeMap);
        stringRedisTemplate.opsForValue().set(projectRunningKey, JsonUtil.beanToJson(projectMessageDTO));
        //去重，之后发送消息的时候会补全指标，如果不去重的话会出现多个场景重复关联多个指标
        Set<ScenePO> scenePOSet = new HashSet<>(scenePOList);
        // -------------------------------- 2 查询模型 --------------------------------
        //2-1 根据车辆配置id vehicleConfigId, 获取 模型信息和传感器信息
        VehiclePO vehiclePO = vehicleMapper.selectByVehicleConfigId(vehicleConfigId);   // 车辆
        List<CameraPO> cameraPOList = sensorCameraMapper.selectCameraByVehicleConfigId(vehicleConfigId);    // 摄像头
        List<OgtPO> ogtPOList = sensorOgtMapper.selectOgtByVehicleId(vehicleConfigId); // 完美传感器
        // -------------------------------- 3 发送任务消息 --------------------------------
        projectService.sendTaskMessage(realCurrentParallelism, projectRunningKey, userId, projectId, projectType, videoTime, scenePOSet, vehiclePO, cameraPOList, ogtPOList);
        // -------------------------------- 4 算法导入（一期按单机版做） --------------------------------
        String algorithmDockerImage = projectService.handleAlgorithm(projectId, algorithmId);
        // -------------------------------- 5 创建 pod 开始执行 --------------------------------
        projectService.createPod(projectId, nodeMap, algorithmDockerImage);

    }

    @KafkaListener(groupId = "simulation-resource-scheduler", topics = "${scheduler.stop-topic}")
    @SneakyThrows
    public void stopProject(ConsumerRecord<String, String> stopRecord) {
        log.info("ProjectConsumer--stopManualProject 接收到的项目终止消息为：" + stopRecord);
        //1 读取 kafka 的项目停止信息
        /*
            {
                "projectId": "sadfasdfs",	// 项目 id
                "type": "1",	// 项目类型
            }
         */

        String json = stopRecord.value();
        ObjectMapper objectMapper = new ObjectMapper();
        JsonNode jsonNode = objectMapper.readTree(json);
        String projectId = jsonNode.path("projectId").asText();
        String type = jsonNode.path("type").asText();
        projectService.stopProject(projectId, type);
    }


}
