package com.css.simulation.resource.scheduler.service;

import api.common.pojo.constants.DictConstants;
import api.common.pojo.dto.ProjectMessageDTO;
import api.common.util.*;
import com.css.simulation.resource.scheduler.configuration.docker.DockerConfiguration;
import com.css.simulation.resource.scheduler.configuration.git.GitConfiguration;
import com.css.simulation.resource.scheduler.configuration.kubernetes.KubernetesConfiguration;
import com.css.simulation.resource.scheduler.mapper.*;
import com.css.simulation.resource.scheduler.pojo.po.*;
import com.css.simulation.resource.scheduler.pojo.to.*;
import com.css.simulation.resource.scheduler.util.*;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import io.minio.MinioClient;
import io.minio.errors.*;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.kafka.clients.admin.Admin;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.kafka.core.KafkaTemplate;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.Resource;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.util.*;
import java.util.stream.Collectors;

@Service
@Slf4j
public class ProjectService {

    /*
      algorithm-platform:
      appid: 2af6f44d98104dc5adcbfb49809ff9d5
      secret: db129a741fde1e9f474199dea24f3901
      token-uri: https://open.zoogooy.com.cn/cgi-bin/token/token?grant_type=client_credential
      algorithm-list-uri: https://open.zoogooy.com.cn/cgi-bin/api/icv-algorithm-agg/simulation/evaluation
      algorithm-addr-uri:  https://open.zoogooy.com.cn/cgi-bin/api/icv-algorithm-agg/simulation/download
     */

    @Value("${algorithm-platform.appid}")
    String algorithmPlatformAppid;
    @Value("${algorithm-platform.secret}")
    String algorithmPlatformSecret;
    @Value("${algorithm-platform.token-uri}")
    String algorithmPlatformTokenUri;
    @Value("${algorithm-platform.algorithm-addr-uri}")
    String algorithmPlatformAlgorithmAddrUri;
    @Value("${scheduler.minio-path.project-result}")
    String projectResultPathOfMinio;
    @Value("${scheduler.linux-path.temp}")
    String linuxTempPath;
    @Value("${scheduler.linux-path.pod-template-yaml}")
    String podTemplateYaml;
    @Value("${scheduler.linux-path.pod-yaml-directory}")
    String podYamlDirectory;
    @Value("${minio.bucket-name}")
    String bucketName;
    @Value("${scheduler.host.hostname}")
    String hostname;
    @Value("${scheduler.host.username}")
    String username;
    @Value("${scheduler.host.password}")
    String password;

    @Resource
    StringRedisTemplate stringRedisTemplate;
    @Resource(name = "myKafkaAdmin")
    Admin kafkaAdminClient;
    @Resource
    KafkaTemplate<String, String> kafkaTemplate;
    @Resource
    CloseableHttpClient closeableHttpClient;
    @Resource
    RequestConfig requestConfig;
    @Resource
    MinioClient minioClient;
    @Resource
    TaskMapper taskMapper;
    @Resource
    IndexTemplateMapper indexTemplateMapper;
    @Resource
    IndexMapper indexMapper;
    @Resource
    SceneMapper sceneMapper;
    @Resource
    AlgorithmMapper algorithmMapper;
    @Resource
    ProjectUtil projectUtil;
    @Resource
    DockerConfiguration dockerConfiguration;
    @Resource
    KubernetesConfiguration kubernetesConfiguration;
    @Resource
    GitConfiguration gitConfiguration;


    // -------------------------------- Comment --------------------------------

    /**
     * @param projectMessageDTO 初始接收到的项目启动信息
     * @param projectWaitingKey projectWaitingKey
     * @param projectRunningKey projectRunningKey
     */
    @Transactional
    public void prepare(ProjectMessageDTO projectMessageDTO, String projectWaitingKey, String projectRunningKey) {
        String projectId = projectMessageDTO.getProjectId();

        //2 将 redis 中该项目旧的信息则直接删除（包括 waitingKey)
        RedisUtil.deleteByPrefix(stringRedisTemplate, projectWaitingKey);
        RedisUtil.deleteByPrefix(stringRedisTemplate, projectRunningKey);


        //5 将该 project 下所有旧的任务和指标得分删除。
        taskMapper.deleteByProject(projectId);
        indexMapper.deleteFirstTargetScoreByProjectId(projectId);
        indexMapper.deleteLastTargetScoreByProjectId(projectId);
    }

    @SneakyThrows
    @Transactional
    public List<ScenePO> handlePackage(String projectRunningKey, String projectId, String packageId) {
        String allIndexPrefix = projectRunningKey + ":package:" + packageId + ":all";
        String leafIndexPrefix = projectRunningKey + ":package:" + packageId + ":leaf";

        //1 查询该场景包的所有指标列表存入 redis，包删了无所谓，但要过滤删掉的指标
        List<IndexTemplatePO> allIndexList = indexTemplateMapper.selectByPackageIdIncludeDeleted(packageId);
        stringRedisTemplate.opsForValue().set(allIndexPrefix, JsonUtil.listToJson(allIndexList));

        //2 查询场景包叶子指标
        List<IndexTemplatePO> leafIndexList = allIndexList.stream().filter(index -> StringUtil.isNotEmpty(index.getRuleId())).collect(Collectors.toList());
        stringRedisTemplate.opsForValue().set(leafIndexPrefix, JsonUtil.listToJson(leafIndexList));
        log.info("ProjectService--handlePackage 项目 " + projectId + " 的叶子指标为：" + leafIndexList);
        List<ScenePO> sceneList = new ArrayList<>();
        leafIndexList.forEach(leafIndex -> {
            String naturalIds = leafIndex.getSceneNaturalIds();
            String standardIds = leafIndex.getSceneStatueIds();
            String accidentIds = leafIndex.getSceneTrafficIds();
            String generalizationIds = leafIndex.getSceneGeneralizationIds();
            if (StringUtil.isNotEmpty(naturalIds)) {
                List<String> naturalIdList = new ArrayList<>(Arrays.asList(naturalIds.split(",")));
                sceneList.addAll(sceneMapper.selectNaturalByIdList(naturalIdList));
            }
            if (StringUtil.isNotEmpty(standardIds)) {
                List<String> standardIdList = new ArrayList<>(Arrays.asList(standardIds.split(",")));
                sceneList.addAll(sceneMapper.selectStandardByIdList(standardIdList));
            }
            if (StringUtil.isNotEmpty(accidentIds)) {
                List<String> accidentIdList = new ArrayList<>(Arrays.asList(accidentIds.split(",")));
                sceneList.addAll(sceneMapper.selectAccidentByIdList(accidentIdList));
            }
            if (StringUtil.isNotEmpty(generalizationIds)) {
                List<String> generalizationIdList = new ArrayList<>(Arrays.asList(generalizationIds.split(",")));
                sceneList.addAll(sceneMapper.selectGeneralizationByIdList(generalizationIdList));
            }
        });
        log.info("ProjectService--handlePackage 项目" + projectId + " 共有 " + sceneList.size() + " 个任务！");
        return sceneList;
    }


    /**
     * @param parallelism          并行度
     * @param projectRunningPrefix
     * @param userId
     * @param projectId
     * @param videoTime            视频长度
     * @param scenePOSet
     * @param vehiclePO
     * @param cameraPOList
     * @param ogtPOList
     */
    @SneakyThrows
    public void sendTaskMessage(int parallelism, String projectRunningPrefix, String userId, String projectId, String projectType, Long videoTime, Set<ScenePO> scenePOSet, VehiclePO vehiclePO, List<CameraPO> cameraPOList, List<OgtPO> ogtPOList) {

        final int[] messageNumber = CollectionUtil.createIntArray(0);
        ApacheKafkaUtil.createTopic(kafkaAdminClient, projectId, parallelism, (short) 1);   // 创建主题
        log.info("ProjectService--sendTaskMessage 项目 " + projectId + " 获得的包括的场景信息为：" + scenePOSet);
        for (ScenePO scenePO : scenePOSet) {
            String sceneId = scenePO.getId();
            //3-1 可能会存在多个指标下有同样的场景，所以会查出多个指标，多个指标的场景需要发送多次
            List<String> lastTargetIdList = null;
            if (DictConstants.PROJECT_TYPE_MANUAL.equals(projectType)) {
                lastTargetIdList = indexMapper.selectLeafIndexIdByManualProjectIdAndSceneId(projectId, "%" + sceneId + "%");
            } else if (DictConstants.PROJECT_TYPE_AUTO_SUB.equals(projectType)) {
                lastTargetIdList = indexMapper.selectLeafIndexIdByAutoSubProjectIdAndSceneId(projectId, "%" + sceneId + "%");
            }
            lastTargetIdList.forEach(lastTargetId -> {
                String taskId = StringUtil.getRandomUUID();
                String taskRetryPrefix = projectRunningPrefix + ":task:" + taskId + ":retry";
                String taskMessagePrefix = projectRunningPrefix + ":task:" + taskId + ":message";
                // 设置任务重试次数为 0，方便任务进行最大3次的重试。
                stringRedisTemplate.opsForValue().set(taskRetryPrefix, "0");
                // 保存任务信息
                TaskPO taskPO = TaskPO.builder() // run_start_time 和 run_end_time 不填
                        .id(taskId)
                        .pId(projectId)
                        .sceneId(sceneId)
                        .lastTargetId(lastTargetId)
                        .sceneName(scenePO.getName())
                        .sceneType(scenePO.getType())
                        .runState(DictConstants.TASK_PENDING)
                        .runResultFilePath(projectResultPathOfMinio + projectId + "/" + taskId)
                        .build();
                taskPO.setCreateTime(TimeUtil.getNowForMysql());
                taskPO.setCreateUserId(userId);
                taskPO.setModifyTime(TimeUtil.getNowForMysql());
                taskPO.setModifyUserId(userId);
                taskPO.setModifyTime(TimeUtil.getNowForMysql());
                taskPO.setIsDeleted("0");
                taskMapper.insert(taskPO);
                // 下载 xodr 和 osgb 供仿真后生成 xosc
                String scenarioOdr = scenePO.getScenarioOdr();
                String scenarioOsgb = scenePO.getScenarioOsgb();
                String[] splitXodr = scenarioOdr.split("/");
                String xodrName = splitXodr[splitXodr.length - 1];
                String[] splitOsgb = scenarioOsgb.split("/");
                String osgbName = splitOsgb[splitOsgb.length - 1];
                try {
                    MinioUtil.downloadToFile(minioClient, bucketName, scenarioOdr, linuxTempPath + "video/" + projectId + "/" + taskId + "/" + xodrName);
                    MinioUtil.downloadToFile(minioClient, bucketName, scenarioOdr, linuxTempPath + "video/" + projectId + "/" + taskId + "/" + osgbName);
                } catch (IOException | ServerException | InsufficientDataException | ErrorResponseException |
                         NoSuchAlgorithmException | InvalidKeyException | InvalidResponseException |
                         XmlParserException | InternalException e) {
                    throw new RuntimeException(e);
                }

                // 组装 task 消息
                TaskTO taskTO = TaskTO.builder()
                        .info(InfoTO.builder()
                                .project_id(taskPO.getPId())
                                .task_id(taskPO.getId())
                                .task_path(taskPO.getRunResultFilePath())
                                .default_time(videoTime)
                                .build())
                        .scenario(ScenarioTO.builder()
                                .scenario_osc(scenePO.getScenarioOsc())
                                .scenario_odr(scenarioOdr)
                                .scenario_osgb(scenarioOsgb)
                                .build())
                        .vehicle(VehicleTO.builder()
                                .model(ModelTO.builder()
                                        .model_label(vehiclePO.getModelLabel())
                                        .build())
                                .dynamics(DynamicsTO.builder()
                                        .dynamics_maxspeed(vehiclePO.getMaxSpeed())
                                        .dynamics_enginepower(vehiclePO.getEnginePower())
                                        .dynamics_maxdecel(vehiclePO.getMaxDeceleration())
                                        .dynamics_maxsteering(vehiclePO.getMaxSteeringAngle())
                                        .dynamics_mass(vehiclePO.getMass())
                                        .dynamics_frontsurfaceeffective(vehiclePO.getFrontSurfaceEffective())
                                        .dynamics_airdragcoefficient(vehiclePO.getAirDragCoefficient())
                                        .dynamics_rollingresistance(vehiclePO.getRollingResistanceCoefficient())
                                        .dynamics_wheeldiameter(vehiclePO.getWheelDiameter())
                                        .dynamics_wheeldrive(vehiclePO.getWheelDrive())
                                        .dynamics_overallefficiency(vehiclePO.getOverallEfficiency())
                                        .dynamics_distfront(vehiclePO.getFrontDistance())
                                        .dynamics_distrear(vehiclePO.getRearDistance())
                                        .dynamics_distleft(vehiclePO.getLeftDistance())
                                        .dynamics_distright(vehiclePO.getRightDistance())
                                        .dynamics_distheight(vehiclePO.getHeightDistance())
                                        .dynamics_wheelbase(vehiclePO.getWheelbase())
                                        .build())
                                .sensors(SensorsTO.builder()   // 根据 vehicleId 查询绑定的传感器列表
                                        .camera(cameraPOList)
                                        .OGT(ogtPOList)
                                        .build())
                                .build())
                        .build();

                //4-4 将对象转成 json
                String taskJson = "";
                try {
                    taskJson = JsonUtil.beanToJson(taskTO);
                } catch (JsonProcessingException e) {
                    e.printStackTrace();
                }
                //4-5 将 projectId 作为 topic 名称，根据 parallelism 分散发送 task 信息到 kafka
                String finalTaskJson = taskJson;
                stringRedisTemplate.opsForValue().set(taskMessagePrefix, finalTaskJson);
                kafkaTemplate.send(projectId, messageNumber[0] % parallelism, "", taskJson).addCallback(success -> {
                    // 消息发送到的topic
                    String topic = success.getRecordMetadata().topic();
                    // 消息发送到的分区
                    int partition = success.getRecordMetadata().partition();
                    // 消息在分区内的offset
                    long offset = success.getRecordMetadata().offset();
                    log.info("------- ProjectConsumer 发送消息成功， 主题 topic 为：" + topic + " 分区 partition 为：" + partition + " 偏移量为：" + offset + " 消息体为：" + finalTaskJson);
                }, failure -> log.error("------- 发送消息失败：" + failure.getMessage()));
                messageNumber[0] = messageNumber[0] + 1;
            });
        }
        log.info("ProjectService--sendTaskMessage 共发送了 " + messageNumber[0] + " 条消息！");

    }


    /**
     * 将 master 节点设置成镜像仓库，导入镜像的同时 commit 到仓库当中，供其他节点 pull
     * {
     * "data" : {
     * "totalElements" : 19,
     * "totalPage" : null,
     * "currentPage" : 0,
     * "size" : 0,
     * "content" : [ {
     * "id" : "ICV20220418170905",
     * "algorithmId" : "2022041813435600064848bbc0e0efb441978d57",
     * "algorithmName" : "SUAN FA2"
     * "description" : "",
     * "algorithmType" : "2"
     * }, {
     * "id" : "ICV20220408165152",
     * "algorithmId" : "202204081650140005916540f84c37764591a5c4",
     * "algorithmName" : "ceshi 33",
     * "description" : "",
     * "algorithmType" : "2"
     * } ],
     * "hasPreviousPage" : false,
     * "hasNextPage" : false
     * },
     * "success" : true,
     * "message" : "ok",
     * "code" : 1,
     * "nowTime" : "2022-04-22 10:14:40"
     * }
     * <p>
     * {
     * "data": "https://sysware-icv-algo-pri.obs.cn-north-4.myhuaweicloud.com:443/1658730835935/jybdc_0712_01.tar?AccessKeyId=0ZFICI4CB04DFHQTPDBF&Expires=1658917777&Signature=bPUT%2FnFwlHcAqOFsskYVPKOdnYs%3D",
     * "success": true,
     * "message": "ok",
     * "code": 1,
     * "nowTime": "2022-07-27 18:14:37"
     * }
     *
     * @param projectId   项目 id
     * @param algorithmId 算法 id
     * @return 镜像名称
     */
    @SneakyThrows
    @Transactional
    public String handleAlgorithm(String projectId, String algorithmId) {
        //4-1 根据算法 id 获取算法文件地址、是否已导入成镜像。
        AlgorithmPO algorithmPO = algorithmMapper.selectById(algorithmId);
        String dockerImage;
        String algorithmDirectoryLinuxTempPath;
        String algorithmTarLinuxTempPath = null;
        if (algorithmPO != null) { // 使用仿真平台自己的算法
            String algorithmCode = algorithmPO.getAlgorithmCode();
            String dockerImport = algorithmPO.getDockerImport();
            dockerImage = dockerConfiguration.getRegistry() + "/algorithm_" + algorithmCode + ":latest";
            if ("1".equals(dockerImport) && StringUtil.isNotEmpty(dockerImport)) { // 已经导入同时存在镜像名称
                return algorithmPO.getDockerImage();
            } else if (dockerImport == null || "0".equals(dockerImport)) {  // 未导入
                String uploadMode = algorithmPO.getUploadMode();
                if (DictConstants.ALGORITHM_UPLOAD_MODE_FILE.equals(uploadMode)) {
                    algorithmTarLinuxTempPath = linuxTempPath + "algorithm-file/" + algorithmCode + "/" + algorithmCode + ".tar";
                    String minioPath = algorithmPO.getMinioPath();
                    MinioUtil.downloadToFile(minioClient, bucketName, minioPath, algorithmTarLinuxTempPath);    // 下载算法文件到本地
                } else if (DictConstants.ALGORITHM_UPLOAD_MODE_GIT.equals(uploadMode)) {
                    algorithmDirectoryLinuxTempPath = linuxTempPath + "algorithm-git/" + algorithmCode + "/";
                    String gitUrl = algorithmPO.getGitUrl().replace(gitConfiguration.getName(), gitConfiguration.getUrl());
                    String gitUserName = algorithmPO.getGitUserName();
                    String gitPassword = algorithmPO.getGitPassword();
                    GitUtil.clone(gitUrl, gitUserName, gitPassword, algorithmDirectoryLinuxTempPath); // 下载算法文件到目录
                    for (String filename : Objects.requireNonNull(new File(algorithmDirectoryLinuxTempPath).list())) {
                        if (filename.endsWith(".tar")) {
                            algorithmTarLinuxTempPath = algorithmDirectoryLinuxTempPath + filename;
                        }
                    }
                } else {
                    throw new RuntimeException("算法 " + algorithmId + " 的 upload_mode 字段数据有误！");
                }
                algorithmMapper.updateDockerImportAndDockerImageById("1", dockerImage, algorithmId);
                if (algorithmTarLinuxTempPath == null) {
                    throw new RuntimeException("ProjectService--handleAlgorithm 算法" + algorithmId + "下载失败。");
                }
                LinuxUtil.execute("docker import " + algorithmTarLinuxTempPath + " " + dockerImage);
                LinuxUtil.execute("docker push " + dockerImage);
                FileUtil.rm(algorithmTarLinuxTempPath);
                log.info("ProjectService--handleAlgorithm 已删除算法临时文件：" + algorithmTarLinuxTempPath);
            } else {
                throw new RuntimeException("算法 " + algorithmId + " 的 mysql 数据有误！");
            }
        } else { // 从索为算法平台下载算法
            log.info("ProjectService--handleAlgorithm 项目" + projectId + "需要使用索为平台算法 " + algorithmId);
            algorithmTarLinuxTempPath = linuxTempPath + "algorithm/" + algorithmId + ".tar";
            String dockerImageWithoutVersion = dockerConfiguration.getRegistry() + "/algorithm_" + algorithmId;
            dockerImage = dockerImageWithoutVersion + ":latest";
            //0 查看算法是否已经导入
            if (!projectUtil.isImported(dockerImageWithoutVersion)) {
                //1 获取 token
                String tokenUrl = algorithmPlatformTokenUri + "?grant_type=client_credential&appid=" + algorithmPlatformAppid + "&secret=" + algorithmPlatformSecret;
                String tokenJson = HttpUtil.get(closeableHttpClient, requestConfig, tokenUrl);
                String token = new ObjectMapper().readTree(tokenJson).path("data").path("access_token").asText();
                //2 获取 下载地址
                String downloadUrl = algorithmPlatformAlgorithmAddrUri + "?access_token=" + token + "&id=" + algorithmId;
                //3 下载算法包
                String downloadUrlJson = HttpUtil.get(closeableHttpClient, requestConfig, downloadUrl);
                String tempDownloadUrl = new ObjectMapper().readTree(downloadUrlJson).path("data").asText();
                InputStream inputStream = HttpUtil.getInputStream(closeableHttpClient, requestConfig, tempDownloadUrl);
                FileUtil.writeInputStreamToLocalFile(inputStream, algorithmTarLinuxTempPath);
                // 本地执行 docker import 算法文件成镜像（可改成用 docker-java 操作仓库）
                LinuxUtil.execute("docker import " + algorithmTarLinuxTempPath + " " + dockerImage);
                LinuxUtil.execute("docker push " + dockerImage);
                FileUtil.rm(algorithmTarLinuxTempPath);
                log.info("ProjectService--handleAlgorithm 已删除算法临时文件：" + algorithmTarLinuxTempPath);
            } else {
                log.info("ProjectService--handleAlgorithm 算法镜像" + dockerImageWithoutVersion + "已导入。");
            }
        }

//        LinuxUtil.execute("docker tag " + algorithmTarLinuxTempPath + " " + dockerImage);   // 标记镜像名称
//        LinuxUtil.execute("docker login " + algorithmTarLinuxTempPath + " " + dockerImage); // 登录 harbor
//        LinuxUtil.execute("docker push " + algorithmTarLinuxTempPath + " " + dockerImage);  // 推送镜像到仓库
        log.info("ProjectService--handleAlgorithm 项目 " + projectId + " 使用的算法镜像为：" + dockerImage);
        return dockerImage;
    }

    /**
     * 运行 pod
     *
     * @param projectId            项目id
     * @param nodeMap              并行度
     * @param algorithmDockerImage 算法镜像
     */
    @SneakyThrows
    public void createPod(String projectId, Map<String, Integer> nodeMap, String algorithmDockerImage) {
        String podString = FileUtil.read(new File(podTemplateYaml));
        String replace0 = podString.replace("vtd-container", "vtd-" + projectId);
        String replace1 = replace0.replace("vtd-image", kubernetesConfiguration.getVtdImage());
        String replace2 = replace1.replace("algorithm-container", "algorithm-" + projectId);
        String replace3 = replace2.replace("algorithm-image", algorithmDockerImage);
        String podTemplateStringOfProject = replace3.replace("kafkaTopic", projectId);     // 消息主题名称为 projectId
        String podTemplateFileNameOfProject = "project-" + projectId + ".yaml";     // 模板文件名称
        FileUtil.writeStringToLocalFile(podTemplateStringOfProject, podYamlDirectory + podTemplateFileNameOfProject);

        nodeMap.forEach((nodeName, parallelism) -> {
            for (int i = 0; i < parallelism; i++) {
                String podName = projectUtil.getRandomPodName(projectId);   // 生成 podName
                String tempReplace4 = podTemplateStringOfProject.replace("pod-name", podName); // pod 名称包括 projectId 和 随机字符串
                String tempReplace5 = tempReplace4.replace("namespace-name", kubernetesConfiguration.getNamespace()); // pod 名称包括 projectId 和 随机字符串
                String tempPodString = tempReplace5.replace("node-name", nodeName);     // 指定 pod 运行节点
//                log.info("ProjectService--createPod 在节点 " + nodeName + " 开始执行 pod：" + tempPodString);
                log.info("ProjectService--createPod 在节点 " + nodeName + " 开始执行 pod。");
                projectUtil.createPod(nodeName, podName, tempPodString);
            }
        });
    }


    /**
     * @param projectId   手动项目 id 或自动项目子id
     * @param projectType 项目类型
     */
    @SneakyThrows
    public void stopProject(String projectId, String projectType) {

        //1 删除项目所有任务
        taskMapper.deleteByProject(projectId);

        //2 根据 pod 前缀删除所有 pod
        String podPrefix = "project-" + projectId;
        Set<String> nodeOfPodKeySet = RedisUtil.getKeySetByPrefix(stringRedisTemplate, "pod:" + podPrefix);
        for (String nodeOfPodKey : nodeOfPodKeySet) {
            String podName = nodeOfPodKey.split(":")[1];
            String nodeName = projectUtil.getNodeNameOfPod(podName);
            // 删除 pod
            projectUtil.deletePod(podName);
            // 节点并行度加一
            projectUtil.addOneParallelismToNode(nodeName);
        }

        //3 其他 redis key
        PrefixTO redisPrefix = projectUtil.getRedisPrefixByProjectIdAndProjectType(projectId, projectType);
        RedisUtil.deleteByPrefix(stringRedisTemplate, redisPrefix.getProjectRunningKey());
        RedisUtil.deleteByPrefix(stringRedisTemplate, redisPrefix.getProjectWaitingKey());
    }


}
